#!/usr/bin/perl -w -- -d:ptkdb
#
# rsa_test.pl
#
# this script tests all classes and it's methods
# which are required by nabou for public-key crypto.
#
# Jus run the script, if it ends up with:
# "Done[14 of 14 tests were successful]."
# then everything is fine.
#
# If not, install the required module and try again.
#
# $Id: rsa_test.pl,v 1.1 2001/06/10 03:59:58 tom Exp tom $
#
# Copyright 2000 (c) Thomas Linden.
# All rights reserved.
#
# This program is published under the terms
# of the GPL. You may redistribute or modify
# the program as you wish.
# The author of the program gives absolutely
# no warranty for damages caused by this
# program. Use it at your own risk.
#
# Of course, you can email me, if you encounter
# any problems or if you find another bug :-)
#
# Thomas Linden <tom@daemon.de>




use strict;

BEGIN {
print "Loading modules:\n\n";
print "use Crypt::OpenSSL::RSA\n";
use Crypt::OpenSSL::RSA;

print "use Crypt::CBC\n";
use Crypt::CBC;

print "use MIME::Base64\n";
use MIME::Base64;

print "use Crypt::Primes qw(maurer)\n";
use Crypt::Primes qw(maurer);
}

my $suck = 4;

my $plain = "  This is the clear text message.
  This is the clear text message.";

my $rsa_gen = new Crypt::OpenSSL::RSA;
$suck++;


print "\nGenerating prime 1024 bit\n";
my $prime = maurer (Size => 1024, Verbosity => 1);
$suck++;

print "\nGenerating 1024 bit keys\n";
$rsa_gen->generate_key(1024, $prime);
my $private = $rsa_gen->get_private_key_string();
my $public  = $rsa_gen->get_public_key_string();
$suck++;



$private =~ s/^.*(-----BEGIN.*KEY-----).*$/$1\n/s;
$public  =~ s/^.*(-----BEGIN.*KEY-----).*$/$1\n/s;


my $rsa = new Crypt::OpenSSL::RSA;

print "Generated keys: \n" . $private . "\n" . $public . "\n";


print "Encrypting public key using pass-phrase \"foo bar\" with CBC::Blowfish\n";
my $cypher = new Crypt::CBC("foo bar", 'Blowfish');
$suck++;
my $cr_public = $public;
$cr_public  =~ s/^.*(-----BEGIN RSA PUBLIC KEY-----)\n(.*)\n(-----END RSA PUBLIC KEY-----).*$/$2/s;
my $public_cypher = encode_base64($cypher->encrypt($cr_public));
$suck++;

$cr_public = "-----BEGIN ENCRYPTED RSA PUBLIC KEY-----\n"
           . $public_cypher
           . "-----END ENCRYPTED RSA PUBLIC KEY-----\n";

print "Encrypted public key: \n" . $cr_public . "\n";


print "Encrypting the following text: \n" . $plain . "\n" . "using the public key\n";
$rsa->load_public_key($public);
$suck++;
my $ciphertext = $rsa->encrypt($plain) or die "Invalid key: $!\n";
$suck++;



print "Encrypted text:\n" . encode_base64($ciphertext) . "\n";


print "De-crypting message using private key\n";
$rsa->load_private_key($private);
$suck++;
my $plaintext = $rsa->decrypt($ciphertext) or die "Invalid key: $!\n";
$suck++;


print "De-crypted text:\n" . $plaintext . "\n";

if ($plaintext eq $plain) {
  $suck++;
  print "De-crypted text equals original message text.\n\n"
}

print "Done[$suck of 14 tests were successful].\n" if($suck == 14);

exit;
